# pylint: disable=protected-access
"""封装部分最常用发送操作
使用的前提:
    已经通过.iotbot.json配置好地址和端口 或者 地址端口保持默认值,即127.0.0.1:8888
"""
import sys
from typing import Union

from .action import Action
from .exceptions import ContextTypeError
from .model import FriendMsg, GroupMsg
from .utils import file_to_base64


def Text(text: str, at=False):
    """发送文字 经支持群消息和好友消息接收函数内调用
    :param text: 文字内容
    :param at:是否艾特发送该消息的用户
    """
    text = str(text)
    # 查找消息上下文 `ctx`变量
    ctx = None
    f = sys._getframe()
    upper = f.f_back
    upper_locals = upper.f_locals
    if 'ctx' in upper_locals and isinstance(upper_locals['ctx'], (FriendMsg, GroupMsg)):
        ctx = upper_locals['ctx']
    else:
        for v in upper_locals.values():
            if isinstance(v, (GroupMsg, FriendMsg)):
                ctx = v
                break
    if ctx is None:
        raise ContextTypeError('经支持群消息和好友消息接收函数内调用')

    action = Action(ctx.CurrentQQ)

    if isinstance(ctx, GroupMsg):
        return action.send_group_text_msg(
            ctx.FromGroupId, content=text, atUser=ctx.FromUserId if at else 0
        )
    if isinstance(ctx, FriendMsg):
        if ctx.TempUin:
            return action.send_private_text_msg(
                toUser=ctx.FromUin, content=text, groupid=ctx.TempUin
            )
        else:
            return action.send_friend_text_msg(ctx.FromUin, text)
    return None


def Picture(pic_url='', pic_base64='', pic_path='', content=''):
    """发送图片 经支持群消息和好友消息接收函数内调用
    :param pic_url: 图片链接
    :param pic_base64: 图片base64编码
    :param pic_path: 图片文件路径
    :param content: 包含的文字消息

    pic_url, pic_base64, pic_path必须给定一项
    """
    assert any([pic_url, pic_base64, pic_path]), '必须给定一项'

    ctx = None
    f = sys._getframe()
    upper = f.f_back
    upper_locals = upper.f_locals
    if 'ctx' in upper_locals and isinstance(upper_locals['ctx'], (FriendMsg, GroupMsg)):
        ctx = upper_locals['ctx']
    else:
        for v in upper_locals.values():
            if isinstance(v, (FriendMsg, GroupMsg)):
                ctx = v
                break
    if ctx is None:
        raise ContextTypeError('经支持群消息和好友消息接收函数内调用')

    action = Action(ctx.CurrentQQ)

    if isinstance(ctx, GroupMsg):
        if pic_url:
            return action.send_group_pic_msg(
                ctx.FromGroupId, picUrl=pic_url, content=content
            )
        elif pic_base64:
            return action.send_group_pic_msg(
                ctx.FromGroupId, picBase64Buf=pic_base64, content=content
            )
        elif pic_path:
            return action.send_group_pic_msg(
                ctx.FromGroupId, picBase64Buf=file_to_base64(pic_path), content=content
            )
    if isinstance(ctx, FriendMsg):
        if pic_url:
            if ctx.TempUin:
                return action.send_private_pic_msg(
                    toUser=ctx.FromUin,
                    groupid=ctx.TempUin,
                    picUrl=pic_url,
                    content=content,
                )
            else:
                return action.send_friend_pic_msg(
                    ctx.FromUin, picUrl=pic_url, content=content
                )
        elif pic_base64:
            if ctx.TempUin:
                return action.send_private_pic_msg(
                    toUser=ctx.FromUin,
                    groupid=ctx.TempUin,
                    picBase64Buf=pic_base64,
                    content=content,
                )
            else:
                return action.send_friend_pic_msg(
                    ctx.FromUin, picBase64Buf=pic_base64, content=content
                )
        elif pic_path:
            if ctx.TempUin:
                return action.send_private_pic_msg(
                    toUser=ctx.FromUin,
                    groupid=ctx.TempUin,
                    picBase64Buf=file_to_base64(pic_path),
                    content=content,
                )
            else:
                return action.send_friend_pic_msg(
                    ctx.FromUin, picBase64Buf=file_to_base64(pic_path), content=content
                )
    return None


def Voice(voice_url='', voice_base64='', voice_path=''):
    """发送语音 经支持群消息和好友消息接收函数内调用
    :param voice_url: 语音链接
    :param voice_base64: 语音base64编码
    :param voice_path: 语音文件路径

    voice_url, voice_base64, voice_path必须给定一项
    """
    assert any([voice_url, voice_base64, voice_path]), '必须给定一项'

    ctx = None
    f = sys._getframe()
    upper = f.f_back
    upper_locals = upper.f_locals
    if 'ctx' in upper_locals and isinstance(upper_locals['ctx'], (FriendMsg, GroupMsg)):
        ctx = upper_locals['ctx']
    else:
        for v in upper_locals.values():
            if isinstance(v, (GroupMsg, FriendMsg)):
                ctx = v
                break
    if ctx is None:
        raise ContextTypeError('经支持群消息和好友消息接收函数内调用')

    action = Action(ctx.CurrentQQ)

    if isinstance(ctx, GroupMsg):
        if voice_url:
            return action.send_group_voice_msg(ctx.FromGroupId, voiceUrl=voice_url)
        elif voice_base64:
            return action.send_group_voice_msg(
                ctx.FromGroupId, voiceBase64Buf=voice_base64
            )
        elif voice_path:
            return action.send_group_voice_msg(
                ctx.FromGroupId, voiceBase64Buf=file_to_base64(voice_path)
            )
    if isinstance(ctx, FriendMsg):
        if voice_url:
            if ctx.TempUin:
                return action.send_private_voice_msg(
                    toUser=ctx.FromUin, groupid=ctx.TempUin, voiceUrl=voice_url
                )
            else:
                return action.send_friend_voice_msg(ctx.FromUin, voiceUrl=voice_url)
        elif voice_base64:
            if ctx.TempUin:
                return action.send_private_voice_msg(
                    toUser=ctx.FromUin, groupid=ctx.TempUin, voiceBase64Buf=voice_base64
                )
            else:
                return action.send_friend_voice_msg(
                    ctx.FromUin, voiceBase64Buf=voice_base64
                )
        elif voice_path:
            if ctx.TempUin:
                return action.send_private_voice_msg(
                    toUser=ctx.FromUin,
                    groupid=ctx.TempUin,
                    voiceBase64Buf=file_to_base64(voice_path),
                )
            else:
                return action.send_friend_voice_msg(
                    ctx.FromUin, voiceBase64Buf=file_to_base64(voice_path)
                )
    return None


def Send(
    ctx: Union[FriendMsg, GroupMsg],
    *,
    text: str = '',
    pic_url: str = '',
    pic_base64: str = '',
    pic_path: str = '',
    voice_url: str = '',
    voice_base64: str = '',
    voice_path: str = ''
):
    """
    根据给定参数自动选择发送方式, 支持对群聊、好友(包括私聊)
    所有参数均为可选参数
    图片和语音支持的三种参数类型，分别三选一，如果传入多个则优先级为: url > base64 > path
    """
    assert isinstance(ctx, (FriendMsg, GroupMsg))
    if text:
        Text(text)
    if any((pic_url, pic_base64, pic_path)):
        Picture(pic_url, pic_base64, pic_path)
    if any((voice_url, voice_base64, voice_path)):
        Voice(voice_url, voice_base64, voice_path)



"""一些常用的方法

Tips: 如果开启队列，请将`action`定义为全局变量!,最重要的一点，开启队列方法都没有返回值，
    所以对于获取信息的api，千万不能用这个模式

对于发送语音，图片的方法，建议将timeout设置很短，因为暂时发现这类请求因为需要文件上传操作，
响应时间会较长，而且目前来看，如果文件较大导致上传时间太长，IOTBOT端会报错, IOTBOT响应的结果一定是错误的,
不过发送去的操作是能正常完成的。
"""
import functools
import queue
import re
import threading
import time
import traceback
from typing import Callable, Generator, List, Union

import requests
from requests.exceptions import Timeout

from . import macro
from .client import IOTBOT
from .config import config
from .logger import logger
from .utils import check_schema

try:
    import ujson as json
except Exception:
    import json


class _Task:
    def __init__(self, target: Callable, args: tuple = None, callback: Callable = None):
        args = args or tuple()
        self.target = functools.partial(target, *args)
        functools.update_wrapper(self.target, target)
        self.callback = callback


class _SendThread(threading.Thread):
    def __init__(self, delay=1.1):
        super().__init__()
        self.tasks = queue.Queue(maxsize=-1)
        self.running = False
        self.delay = delay
        self.last_send_time = time.time()

    def run(self):
        self.running = True
        while True:
            try:
                # 因为重载(importlib.relaod)之后，线程仍会在后台运行
                # 暂时使用超时跳出线程
                # 线程停了之后，被重载后，是不是会被gc??? 0.o
                task: _Task = self.tasks.get(timeout=30 * 60)  # 30min
            except queue.Empty:
                self.running = False
                break
            else:
                should_wait = self.delay - (time.time() - self.last_send_time)
                if should_wait > 0:
                    time.sleep(should_wait)
                try:
                    ret = task.target()
                    if task.callback is not None:
                        task.callback(ret)
                except Exception:
                    logger.exception('Action发送线程出错')
                finally:
                    self.last_send_time = time.time()

    def start(self):
        # 强改内部方法以允许重复执行start方法, 暂时不知道这样做有什么后果
        if not self.running:
            self._started.is_set = lambda: False
        else:
            self._started.is_set = lambda: True
        super().start()

    def put_task(self, task: _Task):
        assert isinstance(task, _Task)
        self.tasks.put(task)
        if not self.running:
            self.start()


class Action:  # pylint:disable=too-many-instance-attributes
    """
    :param qq_or_bot: qq号或者机器人实例(`IOTBOT`)
    :param queue: 是否开启队列，开启后任务将按顺序发送并延时指定时间，此参数与`queue_delay`对应
                  启用后，发送方法`没有返回值`
    :param queue_delay: 与`队列`对应, 开启队列时发送每条消息间的延时, 保持默认即可
    :param timeout: 等待IOTBOT响应时间和发送请求的延时
    :param api_path: 方法路径
    :param port: 端口
    :param host: ip
    """

    def __init__(
        self,
        qq_or_bot: Union[int, IOTBOT] = None,
        queue: bool = False,
        queue_delay: Union[int, float] = 1.1,
        timeout: int = 15,
        api_path: str = '/v1/LuaApiCaller',
        port: int = 8888,
        host: str = 'http://127.0.0.1',
    ):
        self.timeout = timeout
        self.api_path = api_path
        self.port = config.port or port
        self.host = config.host or check_schema(host)
        if isinstance(qq_or_bot, IOTBOT):
            self.bind_bot(qq_or_bot)
        else:
            self.qq = int(qq_or_bot)

        self.s = requests.Session()

        # 任务队列
        self._use_queue = queue
        self._send_thread = _SendThread(queue_delay)
        self._send_thread.setDaemon(True)

    def bind_bot(self, bot: IOTBOT):
        """绑定机器人"""
        self.qq = bot.qq[0]
        self.port = bot.port
        self.host = bot.host

    def send_friend_text_msg(
        self, toUser: int, content: str, timeout=5, **kwargs
    ) -> dict:
        """发送好友文本消息"""
        data = {
            "toUser": toUser,
            "sendToType": 1,
            "sendMsgType": "TextMsg",
            "content": content,
            "groupid": 0,
            "replayInfo": None,
        }
        return self.baseSender('POST', 'SendMsg', data, timeout, **kwargs)

    def get_user_list(self, timeout=5, **kwargs) -> dict:
        """获取好友列表"""
        return self.baseSender(
            'post', 'GetQQUserList', {"StartIndex": 0}, timeout=timeout, **kwargs
        )

    def send_friend_voice_msg(
        self, toUser, voiceUrl='', voiceBase64Buf='', timeout=5, **kwargs
    ) -> dict:
        """发送好友语音消息"""
        data = {
            "toUser": toUser,
            "sendToType": 1,
            "sendMsgType": "VoiceMsg",
            "content": "",
            "groupid": 0,
            "voiceUrl": voiceUrl,
            "voiceBase64Buf": voiceBase64Buf,
        }
        return self.baseSender('POST', 'SendMsg', data, timeout, **kwargs)

    def send_friend_pic_msg(
        self,
        toUser,
        content='',
        picUrl='',
        picBase64Buf='',
        fileMd5: List[str] = None,
        flashPic=False,
        timeout=5,
        **kwargs,
    ) -> dict:
        """发送好友图片消息"""
        data = {
            "toUser": toUser,
            "sendToType": 1,
            "sendMsgType": "PicMsg",
            "content": content,
            "groupid": 0,
            "picUrl": picUrl,
            "picBase64Buf": picBase64Buf,
            "fileMd5": fileMd5,
            "flashPic": flashPic,
        }
        return self.baseSender('POST', 'SendMsg', data, timeout, **kwargs)

    def send_group_text_msg(
        self,
        toUser: int,
        content='',
        atUser: Union[int, List[int]] = 0,
        timeout=5,
        **kwargs,
    ) -> dict:
        """发送群文字消息"""
        if atUser != 0:
            content = macro.atUser(atUser) + content
        data = {
            "toUser": toUser,
            "sendToType": 2,
            "sendMsgType": "TextMsg",
            "content": content,
            "groupid": 0,
        }
        return self.baseSender('POST', 'SendMsg', data, timeout, **kwargs)

    def send_group_voice_msg(
        self, toUser, voiceUrl='', voiceBase64Buf='', timeout=5, **kwargs
    ) -> dict:
        """发送群语音"""
        data = {
            "toUser": toUser,
            "sendToType": 2,
            "sendMsgType": "VoiceMsg",
            "content": '',
            "groupid": 0,
            "voiceUrl": voiceUrl,
            "voiceBase64Buf": voiceBase64Buf,
        }
        return self.baseSender('POST', 'SendMsg', data, timeout, **kwargs)

    def send_group_pic_msg(
        self,
        toUser: int,
        picUrl='',
        flashPic=False,
        atUser: Union[int, List[int]] = 0,
        content='',
        picBase64Buf='',
        fileMd5: List[str] = None,  # 多图
        timeout=5,
        **kwargs,
    ) -> dict:
        """发送群图片
        Tips:
            [秀图id] 各id对应效果
            40000 秀图  40001 幻影  40002 抖动 40003 生日
            40004 爱你  40005 征友  40006 无(只显示大图无特效)

            [PICFLAG] 改变图文消息顺序
        """
        if atUser != 0:
            content = macro.atUser(atUser) + content
        data = {
            "toUser": toUser,
            "sendToType": 2,
            "sendMsgType": "PicMsg",
            "content": content,
            "groupid": 0,
            "picUrl": picUrl,
            "picBase64Buf": picBase64Buf,
            "fileMd5": fileMd5,
            "flashPic": flashPic,
        }
        return self.baseSender('POST', 'SendMsg', data, timeout, **kwargs)

    def send_private_text_msg(
        self, toUser: int, content: str, groupid: int, timeout=5, **kwargs
    ) -> dict:
        """发送私聊文字消息"""
        data = {
            "toUser": toUser,
            "sendToType": 3,
            "sendMsgType": "TextMsg",
            "content": content,
            "groupid": groupid,
        }
        return self.baseSender('POST', 'SendMsg', data, timeout, **kwargs)

    def send_private_voice_msg(
        self, toUser: int, groupid, voiceUrl='', voiceBase64Buf='', timeout=5, **kwargs
    ) -> dict:
        """发送私聊语音"""
        data = {
            "toUser": toUser,
            "sendToType": 3,
            "sendMsgType": "VoiceMsg",
            "content": "",
            "groupid": groupid,
            "voiceUrl": voiceUrl,
            "voiceBase64Buf": voiceBase64Buf,
        }
        return self.baseSender('POST', 'SendMsg', data, timeout, **kwargs)

    def send_private_pic_msg(
        self,
        toUser,
        groupid,
        picUrl='',
        picBase64Buf='',
        content='',
        fileMd5: List[str] = None,
        timeout=10,
        **kwargs,
    ) -> dict:
        """发送私聊图片"""
        data = {
            "toUser": toUser,
            "sendToType": 3,
            "sendMsgType": "PicMsg",
            "content": content,
            "groupid": groupid,
            "picUrl": picUrl,
            "picBase64Buf": picBase64Buf,
            "fileMd5": fileMd5,
        }
        return self.baseSender('POST', 'SendMsg', data, timeout, **kwargs)

    def send_group_json_msg(self, toUser: int, content='', timeout=5, **kwargs) -> dict:
        """发送群Json类型信息
        :param content: 可以为json文本，或者字典类型
        """
        if isinstance(content, dict):
            content = json.dumps(content)
        data = {
            "toUser": toUser,
            "sendToType": 2,
            "sendMsgType": "JsonMsg",
            "content": content,
            "groupid": 0,
        }
        return self.baseSender('POST', 'SendMsg', data, timeout, **kwargs)

    def send_group_xml_msg(self, toUser: int, content='', timeout=5, **kwargs) -> dict:
        """发送群Xml类型信息"""
        data = {
            "toUser": toUser,
            "sendToType": 2,
            "sendMsgType": "XmlMsg",
            "content": content,
            "groupid": 0,
        }
        return self.baseSender('POST', 'SendMsg', data, timeout, **kwargs)

    def revoke_msg(
        self, groupid: int, msgseq: int, msgrandom: int, type_=1, timeout=5, **kwargs
    ) -> dict:
        """撤回消息
        :param type_: 1: RevokeMsg | 2: PbMessageSvc.PbMsgWithDraw
        """
        funcname = 'RevokeMsg' if type_ == 1 else 'PbMessageSvc.PbMsgWithDraw'
        data = {"GroupID": groupid, "MsgSeq": msgseq, "MsgRandom": msgrandom}
        return self.baseSender('POST', funcname, data, timeout, **kwargs)

    def search_group(self, content, page=0, timeout=5, **kwargs) -> dict:
        """搜索群组"""
        return self.baseSender(
            'POST', 'SearchGroup', {"Content": content, "Page": page}, timeout, **kwargs
        )

    def get_user_info(self, userID: int, timeout=5, **kwargs) -> dict:
        '''获取用户信息'''
        return self.baseSender(
            'POST', 'GetUserInfo', {'UserID': userID, 'GroupID': 0}, timeout, **kwargs
        )

    def get_cookies(self, timeout=2, **kwargs) -> dict:
        """获取cookies"""
        return self.baseSender('GET', 'GetUserCook', timeout=timeout, **kwargs)

    def get_group_list(self, timeout=5, **kwargs) -> dict:
        """获取群组列表"""
        return self.baseSender(
            'POST', 'GetGroupList', {"NextToken": ""}, timeout, **kwargs
        )

    def get_group_admin_list(self, groupid: int, timeout=5, **kwargs) -> List[dict]:
        """获取群管理员列表"""
        members = self.get_group_user_list(groupid, timeout, **kwargs)
        return [member for member in members if member['GroupAdmin'] == 1]

    def get_group_all_admin_list(self, groupid: int, timeout=5, **kwargs) -> List[dict]:
        """群管理列表+群主"""
        owner = None
        for group in self.get_group_list(timeout)['TroopList']:
            if group['GroupId'] == groupid:
                owner = group['GroupOwner']
                break
        members = self.get_group_user_list(groupid, timeout, **kwargs)
        return [
            member
            for member in members
            if member['GroupAdmin'] == 1 or member['MemberUin'] == owner
        ]

    def get_group_user_list(
        self, groupid: int, timeout=5, **kwargs
    ) -> Generator[dict, None, None]:
        """获取群成员列表"""
        LastUin = 0
        while True:
            data = self.baseSender(
                'POST',
                'GetGroupUserList',
                {"GroupUin": groupid, "LastUin": LastUin},
                timeout,
                **kwargs,
            )
            LastUin = data['LastUin']  # 上面请求失败会返回空字典。但是这里不处理错误, 必须正常抛出
            for member in data['MemberList']:
                yield member
            if LastUin == 0:
                break
            time.sleep(0.8)

    def set_unique_title(
        self, groupid: int, userid: int, Title: str, timeout=5, **kwargs
    ) -> dict:
        """设置群成员头衔"""
        return self.baseSender(
            'POST',
            'OidbSvc.0x8fc_2',
            {"GroupID": groupid, "UserID": userid, "NewTitle": Title},
            timeout,
            **kwargs,
        )

    def modify_group_card(
        self, userID: int, groupID: int, newNick: str, timeout=5, **kwargs
    ) -> dict:
        """修改群名片
        :params userID: 修改的QQ号
        :params groupID: 群号
        :params newNick: 新群名片
        """
        data = {'UserID': userID, 'GroupID': groupID, 'NewNick': newNick}
        return self.baseSender('POST', 'ModifyGroupCard', data, timeout, **kwargs)

    def refresh_keys(self, timeout=20) -> bool:
        '''刷新key二次登陆, 成功返回True， 失败返回False'''
        try:
            rep = self.s.get(
                f'{self.host}:{self.port}/v1/RefreshKeys?qq={self.qq}', timeout=timeout
            )
            if rep.json()['Ret'] == 'ok':
                return True
        except Exception:
            pass
        return False

    def get_balance(self) -> dict:
        '''获取QQ钱包余额'''
        # TODO

    #     return self.baseSender('GET', 'GetBalance', timeout=timeout, **kwargs)

    def get_status(self, timeout=20) -> dict:
        '''获取机器人状态'''
        rep = self.s.get('{self.host}:{self.port}/v1/ClusterInfo', timeout=timeout)
        return rep.json()

    def send_single_red_bag(self) -> dict:
        '''发送群/好友红包'''
        # TODO
        # data = {}
        # return self.baseSender('POST', 'SendSingleRed', data, timeout, **kwargs)

    def send_qzone_red_bag(self) -> dict:
        '''发送QQ空间红包'''
        # TODO
        # data = {}
        # return self.baseSender('POST', 'SendQzoneRed', data, timeout, **kwargs)

    def send_transfer(self) -> dict:
        '''支付转账'''
        # TODO
        # data = {}
        # return self.baseSender('POST', 'Transfer', data, timeout, **kwargs)

    def open_red_bag(self, OpenRedBag) -> dict:
        '''打开红包 传入红包数据结构'''
        # TODO

    #     return self.baseSender('POST', 'OpenRedBag', OpenRedBag, timeout, **kwargs)

    def add_friend(
        self,
        userID: int,
        groupID: int,
        content='加个好友!',
        AddFromSource=2004,
        timeout=20,
        **kwargs,
    ) -> dict:
        """添加好友"""
        data = {
            "AddUserUid": userID,
            "FromGroupID": groupID,
            "AddFromSource": AddFromSource,
            "Content": content,
        }
        return self.baseSender('POST', 'AddQQUser', data, timeout, **kwargs)

    def get_friend_file(self, FileID: str, timeout=20, **kwargs) -> dict:
        """获取好友文件下载链接"""
        funcname = 'OfflineFilleHandleSvr.pb_ftn_CMD_REQ_APPLY_DOWNLOAD-1200'
        data = {'FileID': FileID}
        return self.baseSender('POST', funcname, data, timeout, **kwargs)

    def get_group_file(self, groupID: int, FileID: str, timeout=20, **kwargs) -> dict:
        """获取群文件下载链接"""
        funcname = 'OidbSvc.0x6d6_2'
        data = {'FileID': FileID, 'GroupID': groupID}
        return self.baseSender('POST', funcname, data, timeout, **kwargs)

    def set_group_announce(
        self,
        groupID: int,
        Title: str,
        Text: str,
        Pinned=0,
        Type=10,
        timeout=5,
        **kwargs,
    ) -> dict:
        """设置群公告"""
        data = {
            'GroupID': groupID,  # 发布的群号
            "Title": Title,  # 公告标题
            "Text": Text,  # 公告内容
            "Pinned": Pinned,  # 1为置顶,0为普通公告
            "Type": Type,  # 发布类型(10为使用弹窗公告,20为发送给新成员,其他暂未知)
        }
        try:
            res = self.s.post(
                f'{self.host}:{self.port}/v1/Group/Announce?qq={self.qq}',
                data=data,
                timeout=timeout,
                **kwargs,
            )
            return res.json()
        except Exception:
            return {}

    def deal_friend(self, Action: int) -> dict:
        """处理好友请求"""
        # TODO
        # --Action 1 忽略 2 同意 3 拒绝
        # data = {
        #     'Action':Action
        # }
        # return self.baseSender('POST', 'DealFriend', data, timeout, **kwargs)

    def deal_group(self, Action: int) -> dict:
        '''处理群邀请'''
        # TODO
        # --Action 14 忽略 1 同意 21 拒绝
        # data = {
        #     'Action':Action
        # }
        # return self.baseSender('POST', 'AnswerInviteGroup', data, timeout, **kwargs)

    def all_shut_up_on(self, groupid, timeout=20, **kwargs) -> dict:
        """开启全员禁言"""
        return self.baseSender(
            'POST',
            'OidbSvc.0x89a_0',
            {"GroupID": groupid, "Switch": 1},
            timeout,
            **kwargs,
        )

    def all_shut_up_off(self, groupid, timeout=20, **kwargs) -> dict:
        """关闭全员禁言"""
        return self.baseSender(
            'POST',
            'OidbSvc.0x89a_0',
            {"GroupID": groupid, "Switch": 0},
            timeout,
            **kwargs,
        )

    def you_shut_up(self, groupid, userid, shut_time=0, timeout=20, **kwargs) -> dict:
        """群成员禁言"""
        return self.baseSender(
            'POST',
            'OidbSvc.0x570_8',
            {"GroupID": groupid, "ShutUpUserID": userid, "ShutTime": shut_time},
            timeout,
            **kwargs,
        )

    def like(self, userid: int, timeout=10, **kwargs) -> dict:
        """通用点赞"""
        return self.baseSender(
            'POST', 'OidbSvc.0x7e5_4', {"UserID": userid}, timeout, **kwargs
        )

    def like_2(self, userid: int, timeout=10, **kwargs) -> dict:
        """测试赞(这里的测试只是与webapi描述一致)"""
        return self.baseSender('POST', 'QQZan', {"UserID": userid}, timeout, **kwargs)

    def logout(self, flag=False, timeout=5, **kwargs) -> bool:
        """退出QQ
        :param flag:是否删除设备信息文件
        """
        return self.baseSender('POST', 'LogOut', {"Flag": flag}, timeout, **kwargs)

    def set_group_admin(self, groupID: int, userID: int, timeout=10, **kwargs) -> dict:
        '''设置群管理员'''
        return self.baseSender(
            'POST',
            'OidbSvc.0x55c_1',
            {"GroupID": groupID, "UserID": userID, "Flag": 1},
            timeout,
            **kwargs,
        )

    def cancel_group_admin(
        self, groupID: int, userID: int, timeout=10, **kwargs
    ) -> dict:
        '''取消群管理员'''
        return self.baseSender(
            'POST',
            'OidbSvc.0x55c_1',
            {"GroupID": groupID, "UserID": userID, "Flag": 0},
            timeout,
            **kwargs,
        )

    def repost_video_to_group(
        self, groupID: int, forwordBuf: str, timeout=10, **kwargs
    ) -> dict:
        '''转发视频到群聊'''
        data = {
            "toUser": groupID,
            "sendToType": 2,
            "sendMsgType": "ForwordMsg",
            "content": "",
            "groupid": 0,
            "forwordBuf": forwordBuf,
            "forwordField": 19,
        }
        return self.baseSender('POST', 'SendMsg', data, timeout, **kwargs)

    def repost_video_to_friend(
        self, userID: int, forwordBuf: str, timeout=10, **kwargs
    ) -> dict:
        '''转发视频给好友'''
        data = {
            "toUser": userID,
            "sendToType": 1,
            "sendMsgType": "ForwordMsg",
            "content": "",
            "groupid": 0,
            "forwordBuf": forwordBuf,
            "forwordField": 19,
        }
        return self.baseSender('POST', 'SendMsg', data, timeout, **kwargs)

    def get_login_qrcode(self) -> str:
        '''返回登录二维码的base64'''
        try:
            resp = self.s.get(
                '{}:{}/v1/Login/GetQRcode'.format(self.host, self.port), timeout=10
            )
        except Exception as e:
            logger.error('http请求错误 %s' % str(e))
        else:
            try:
                return re.findall(r'"data:image/png;base64,(.*?)"', resp.text)[0]
            except IndexError:
                logger.error('base64获取失败')
        return ''

    def get_schedules(self, **kwargs) -> dict:
        """获取定时任务总数
        response {"Crons": "任务总数0\n", "Ret": 0}
        """
        return self.baseSender('GET', 'GetCrons', **kwargs)

    def add_schedules(self, Sepc: str, FileName: str, FuncName: str, **kwargs) -> dict:
        """添加定时任务
        :param sepc: cron表达式
        :param FileName: 执行的lua文件名
        :param FuncName: 指定在该lua文件下的方法
        """
        data = {
            "QQ": str(self.qq),  # 执行任务的机器人
            "Sepc": Sepc,  # cron表达式 每5秒执行一次
            "FileName": FileName,  # 执行的lua文件名
            "FuncName": FuncName,  # 执行的lua文件名下的TaskTwo方法名
        }
        return self.baseSender('POST', 'AddCrons', data, **kwargs)

    def del_schedules(self, TaskID: int, **kwargs) -> dict:
        """删除定时任务
        :param TaskID: 任务ID
        """
        return self.baseSender('POST', 'DelCrons', {'TaskID': TaskID}, **kwargs)

    def baseSender(
        self,
        method: str,
        funcname: str,
        data: dict = None,
        timeout: int = None,
        api_path: str = None,
        iot_timeout: int = None,
        bot_qq: int = None,
        **kwargs,
    ) -> dict:
        """
        :param method: 请求方法
        :param funcname: 请求类型
        :param data: post的数据
        :param timeout: 发送请求等待响应的时间
        :param api_path: 默认为/v1/LuaApiCaller
        :param iot_timeout: IOT端处理请求等待的时间
        :param bot_qq: 机器人QQ

        :return: iotbot端返回的json数据(字典)，其他情况一律返回空字典
        """
        job = functools.partial(
            self._baseSender,
            method=method,
            funcname=funcname,
            data=data,
            timeout=timeout,
            api_path=api_path,
            iot_timeout=iot_timeout,
            bot_qq=bot_qq,
        )
        functools.update_wrapper(job, self.baseSender)
        if self._use_queue:
            self._send_thread.put_task(
                _Task(target=job, callback=kwargs.get('callback'))
            )
            return None
        return job()

    def _baseSender(
        self,
        method: str,
        funcname: str,
        data: dict = None,
        timeout: int = None,
        api_path: str = None,
        iot_timeout: int = None,
        bot_qq: int = None,
    ) -> dict:
        params = {
            'funcname': funcname,
            'timeout': iot_timeout or self.timeout,
            'qq': bot_qq or self.qq,
        }
        if data is None:
            data = {}
        try:
            rep = self.s.request(
                method=method,
                url='{}:{}{}'.format(self.host, self.port, api_path or self.api_path),
                headers={'Content-Type': 'application/json'},
                params=params,
                json=data,
                timeout=timeout or self.timeout,
            )
            if rep.status_code != 200:
                logger.error(
                    f'HTTP响应码错误, 请检查地址端口是否正确, \
                             {self.host}:{self.port} => {rep.status_code}'
                )
                return {}
            response = rep.json()
            self._report_response(response)
            return response
        except Exception as e:
            if isinstance(e, Timeout):
                logger.warning('响应超时，但不代表处理未成功, 结果未知!')
            else:
                logger.error(f'出现错误 => {traceback.format_exc()}')
            return {}

    def _report_response(self, response):
        if response is None:
            logger.error(
                '可能是SendMsg返回为null\n'
                '可能的原因：\n'
                '1. 发送消息超过特定版本的长度限制，比如在x86_linux上发送超过215个汉字或645个ASCII\n'
            )
        elif isinstance(response, dict) and 'Ret' in response:
            ret = response['Ret']
            if ret == 0:
                pass
            elif ret == 34:
                logger.error(f'未知错误，跟消息长度似乎无关，可以尝试分段重新发送 => {response}')
            elif ret == 110:
                logger.error(f'发送失败，你已被移出该群，请重新加群 => {response}')
            elif ret == 120:
                logger.error(f'机器人被禁言 => {response}')
            elif ret == 241:
                logger.error(f'消息发送频率过高，对同一个群或好友，建议发消息的最小间隔控制在1100ms以上 => {response}')
            elif ret == 299:
                logger.error(f'超过群发言频率限制 => {response}')
            else:
                logger.error(f'请求发送成功, 但处理失败 => {response}')



# 该功能即是一个内置插件
import traceback

import requests

from . import sugar
from .config import config
from .exceptions import InvalidConfigError
from .model import EventMsg, FriendMsg, GroupMsg


def _check_config():
    if not config.webhook_post_url:
        raise InvalidConfigError('缺少配置项: webhook_post_url')


_check_config()


def receive_group_msg(ctx: GroupMsg):
    try:
        resp = requests.post(
            config.webhook_post_url, json=ctx.message, timeout=config.webhook_timeout
        )
        resp.raise_for_status()
    except Exception:
        print(traceback.format_exc())
    else:
        try:
            data = resp.json()
            assert isinstance(data, dict)
        except Exception:
            pass
        else:
            # 取出所有支持的字段
            # 1. 图片(pic)存在，发送图片消息，此时文字(msg)存在, 则发送图文消息
            # 2. 图片(pic)不存在, 文字(msg)存在，单独发送文字消息
            # 3. 语音(voice)只要存在，则发送
            msg: str = data.get('msg') or ''
            at: bool = bool(data.get('at')) or False  # 只要存在值就判定真
            pic_url: str = data.get('pic_url')
            pic_base64: str = data.get('pic_base64')
            voice_url: str = data.get('voice_url')
            voice_base64: str = data.get('voice_base64')
            if any([pic_url, pic_base64]):  # 图片，纯文字二选一
                sugar.Picture(pic_url=pic_url, pic_base64=pic_base64, content=msg)
            elif msg:
                sugar.Text(msg, at)
            if any([voice_url, voice_base64]):
                sugar.Voice(voice_url=voice_url, voice_base64=voice_base64)
            return None
    return None


def receive_friend_msg(ctx: FriendMsg):
    try:
        resp = requests.post(
            config.webhook_post_url, json=ctx.message, timeout=config.webhook_timeout
        )
        resp.raise_for_status()
    except Exception:
        print(traceback.format_exc())
    else:
        try:
            data = resp.json()
            assert isinstance(data, dict)
        except Exception:
            pass
        else:
            # 取出所有支持的字段
            # 1. 图片(pic)存在，发送图片消息，此时文字(msg)存在, 则发送图文消息
            # 2. 图片(pic)不存在, 文字(msg)存在，单独发送文字消息
            # 3. 语音(voice)只要存在，则发送
            msg: str = data.get('msg') or ''
            at: bool = bool(data.get('at')) or False  # 只要存在值就判定真
            pic_url: str = data.get('pic_url')
            pic_base64: str = data.get('pic_base64')
            voice_url: str = data.get('voice_url')
            voice_base64: str = data.get('voice_base64')
            if any([pic_url, pic_base64]):  # 图片，纯文字二选一
                sugar.Picture(pic_url=pic_url, pic_base64=pic_base64, content=msg)
            elif msg:
                sugar.Text(msg, at)
            if any([voice_url, voice_base64]):
                sugar.Voice(voice_url=voice_url, voice_base64=voice_base64)
            return None
    return None


def receive_events(ctx: EventMsg):
    # 事件消息只上报(懒)
    try:
        requests.post(
            config.webhook_post_url, json=ctx.message, timeout=config.webhook_timeout
        )
    except Exception:
        pass



import functools
import re

from .model import FriendMsg, GroupMsg


def in_content(string: str):
    """
    接受消息content字段含有指定消息时, 不支持事件类型消息
    :param string: 支持正则
    """

    def deco(func):
        def inner(ctx):
            if isinstance(ctx, (GroupMsg, FriendMsg)):
                if re.findall(string, ctx.Content):
                    return func(ctx)
            return None

        return inner

    return deco


def equal_content(string: str):
    """
    content字段与指定消息相等时, 不支持事件类型消息
    """

    def deco(func):
        def inner(ctx):
            if isinstance(ctx, (GroupMsg, FriendMsg)):
                if ctx.Content == string:
                    return func(ctx)
            return None

        return inner

    return deco


def not_botself(func=None):
    """忽略机器人自身的消息"""
    if func is None:
        return functools.partial(not_botself)

    def inner(ctx):
        if isinstance(ctx, (GroupMsg, FriendMsg)):
            if isinstance(ctx, GroupMsg):
                userid = ctx.FromUserId
            else:
                userid = ctx.FromUin
            if userid != ctx.CurrentQQ:
                return func(ctx)
        return None

    return inner


def is_botself(func=None):
    """只要机器人自身的消息"""
    if func is None:
        return functools.partial(not_botself)

    def inner(ctx):
        if isinstance(ctx, (GroupMsg, FriendMsg)):
            if isinstance(ctx, GroupMsg):
                userid = ctx.FromUserId
            else:
                userid = ctx.FromUin
            if userid == ctx.CurrentQQ:
                return func(ctx)
        return None

    return inner


def not_these_users(users: list):  # pylint:disable=W0613
    """不接受这些人的消息
    :param users: qq号列表
    """

    def deco(func):
        def inner(ctx):
            nonlocal users
            if isinstance(ctx, (GroupMsg, FriendMsg)):
                if not hasattr(users, '__iter__'):
                    users = [users]
                if isinstance(ctx, GroupMsg):
                    from_user = ctx.FromUserId
                elif isinstance(ctx, FriendMsg):
                    from_user = ctx.FromUin
                if from_user not in users:
                    return func(ctx)
            return None

        return inner

    return deco


def only_these_users(users: list):  # pylint:disable=W0613
    """仅接受这些人的消息
    :param users: qq号列表
    """

    def deco(func):
        def inner(ctx):
            nonlocal users
            if isinstance(ctx, (GroupMsg, FriendMsg)):
                if not hasattr(users, '__iter__'):
                    users = [users]
                if isinstance(ctx, GroupMsg):
                    from_user = ctx.FromUserId
                elif isinstance(ctx, FriendMsg):
                    from_user = ctx.FromUin
                if from_user in users:
                    return func(ctx)
            return None

        return inner

    return deco


def only_this_msg_type(msg_type: str):
    """仅接受该类型消息
    :param msg_type: TextMsg, PicMsg, AtMsg, ReplyMsg, VoiceMsg之一
    """

    def deco(func):
        def inner(ctx):
            if isinstance(ctx, (GroupMsg, FriendMsg)):
                if ctx.MsgType == msg_type:
                    return func(ctx)
            return None

        return inner

    return deco


def not_these_groups(groups: list):  # pylint:disable=W0613
    """不接受这些群组的消息
    :param groups: 群号列表
    """

    def deco(func):
        def inner(ctx):
            nonlocal groups
            if isinstance(ctx, GroupMsg):
                if not hasattr(groups, '__iter__'):
                    groups = [groups]
                from_group = ctx.FromGroupId
                if from_group not in groups:
                    return func(ctx)
            return None

        return inner

    return deco


def only_these_groups(groups: list):  # pylint:disable=W0613
    """只接受这些群组的消息
    :param groups: 群号列表
    """

    def deco(func):
        def inner(ctx):
            nonlocal groups
            if isinstance(ctx, GroupMsg):
                if not hasattr(groups, '__iter__'):
                    groups = [groups]
                from_group = ctx.FromGroupId
                if from_group in groups:
                    return func(ctx)
            return None

        return inner

    return deco



# pylint: disable=too-many-instance-attributes, super-init-not-called
"""进一步提取消息详细信息的函数"""
import copy
import functools
from typing import List

from .exceptions import ContextTypeError
from .model import EventMsg, FriendMsg, GroupMsg
from .utils import EventNames, MsgTypes

try:
    import ujson as json
except Exception:
    import json


def copy_ctx(f):
    @functools.wraps(f)
    def i(ctx):
        return f(copy.deepcopy(ctx))

    return i


#############################Event begin########################################


class _EventMsg(EventMsg):
    def _carry_properties(self, ctx: EventMsg):
        self.message = ctx.message
        self.CurrentQQ = ctx.CurrentQQ

        self.data = ctx.data

        self.EventName = ctx.EventName
        self.EventData = ctx.EventData
        self.EventMsg = ctx.EventMsg

        self.Content = ctx.Content
        self.FromUin = ctx.FromUin
        self.MsgSeq = ctx.MsgSeq
        self.MsgType = ctx.MsgType
        self.ToUin = ctx.ToUin
        self.RedBaginfo = ctx.RedBaginfo


class _GroupRevokeEventMsg(_EventMsg):
    """群成员撤回消息事件"""

    def __init__(self, ctx: EventMsg):
        event_data = ctx.EventData
        self.AdminUserID: int = event_data.get('AdminUserID')
        self.GroupID: int = event_data.get('GroupID')
        self.MsgRandom: int = event_data.get('MsgRandom')
        self.MsgSeq: int = event_data.get('MsgSeq')
        self.UserID: int = event_data.get('UserID')
        super()._carry_properties(ctx)


class _GroupExitEventMsg(_EventMsg):
    """群成员退出群聊事件"""

    def __init__(self, ctx: EventMsg):
        self.UserID = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


class _GroupJoinEventMsg(_EventMsg):
    """某人进群事件"""

    def __init__(self, ctx: EventMsg):
        e_data = ctx.EventData
        self.InviteUin: int = e_data.get('InviteUin')
        self.UserID: int = e_data.get('UserID')
        self.UserName: str = e_data.get('UserName')
        super()._carry_properties(ctx)


class _FriendRevokeEventMsg(_EventMsg):
    """好友撤回消息事件"""

    def __init__(self, ctx: EventMsg):
        self.MsgSeq = ctx.EventData.get('MsgSeq')
        self.UserID = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


class _FriendDeleteEventMsg(_EventMsg):
    """删除好友事件"""

    def __init__(self, ctx: EventMsg):
        self.UserID: int = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


class _GroupAdminsysnotifyEventMsg(_EventMsg):
    """QQ群系统消息通知(加群申请在这里面"""

    def __init__(self, ctx: EventMsg):
        edata = ctx.EventData
        self.Type: int = edata.get('Type')  # 事件类型
        self.MsgTypeStr: str = edata.get('MsgTypeStr')  # 消息类型
        self.MsgStatusStr: str = edata.get('MsgStatusStr')  # 消息类型状态
        self.Who: int = edata.get('Who')  # 触发消息的对象
        self.WhoName: int = edata.get('WhoName')  # 触发消息的对象昵称
        self.GroupID: int = edata.get('GroupId')  # 来自群
        self.GroupName: str = edata.get('GroupName')  # 群名
        self.ActionUin: int = edata.get('ActionUin')  # 邀请人(处理人)
        self.ActionName: str = edata.get('ActionName')  # 邀请人(处理人)昵称
        self.ActionGroupCard: str = edata.get('ActionGroupCard')  # 邀请人(处理人)群名片
        self.Action: str = edata.get('Action')  # 加群理由 11 agree 14 忽略 12/21 disagree
        self.Content: int = edata.get('Content')
        super()._carry_properties(ctx)


class _GroupShutEventMsg(_EventMsg):
    """群禁言事件"""

    def __init__(self, ctx: EventMsg):
        self.GroupID: int = ctx.EventData.get('GroupID')
        self.ShutTime: int = ctx.EventData.get('ShutTime')
        self.UserID: int = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


class _GroupAdminEventMsg(_EventMsg):
    """管理员变更事件"""

    def __init__(self, ctx: EventMsg):
        self.Flag: int = ctx.EventData.get('Flag')
        self.GroupID: int = ctx.EventData.get('GroupID')
        self.UserID: int = ctx.EventData.get('UserID')
        super()._carry_properties(ctx)


@copy_ctx
def refine_group_revoke_event_msg(ctx: EventMsg) -> _GroupRevokeEventMsg:
    """群成员撤回消息事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_REVOKE:
        return _GroupRevokeEventMsg(ctx)
    return None


@copy_ctx
def refine_group_exit_event_msg(ctx: EventMsg) -> _GroupExitEventMsg:
    """群成员退出群聊事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_EXIT:
        return _GroupExitEventMsg(ctx)
    return None


@copy_ctx
def refine_group_join_event_msg(ctx: EventMsg) -> _GroupJoinEventMsg:
    """某人进群事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_JOIN:
        return _GroupJoinEventMsg(ctx)
    return None


@copy_ctx
def refine_friend_revoke_event_msg(ctx: EventMsg) -> _FriendRevokeEventMsg:
    """好友撤回消息事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_FRIEND_REVOKE:
        return _FriendRevokeEventMsg(ctx)
    return None


@copy_ctx
def refine_friend_delete_event_msg(ctx: EventMsg) -> _FriendDeleteEventMsg:
    """删除好友事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_FRIEND_DELETE:
        return _FriendDeleteEventMsg(ctx)
    return None


@copy_ctx
def refine_group_adminsysnotify_event_msg(
    ctx: EventMsg,
) -> _GroupAdminsysnotifyEventMsg:
    """加群申请"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_ADMINSYSNOTIFY:
        return _GroupAdminsysnotifyEventMsg(ctx)
    return None


@copy_ctx
def refine_group_shut_event_msg(ctx: EventMsg) -> _GroupShutEventMsg:
    """群禁言事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_SHUT:
        return _GroupShutEventMsg(ctx)
    return None


@copy_ctx
def refine_group_admin_event_msg(ctx: EventMsg) -> _GroupAdminEventMsg:
    """管理员变更事件"""
    if not isinstance(ctx, EventMsg):
        raise ContextTypeError('Expected `EventMsg`, but got `%s`' % ctx.__class__)
    if ctx.EventName == EventNames.ON_EVENT_GROUP_ADMIN:
        return _GroupAdminEventMsg(ctx)
    return None


#############################Event end##########################################


#############################Group start########################################
class _GroupMsg(GroupMsg):
    def _carry_properties(self, ctx: GroupMsg):
        self.message = ctx.message
        self.CurrentQQ = ctx.CurrentQQ

        self.data = ctx.data

        self.FromGroupId: int = ctx.FromGroupId
        self.FromGroupName: str = ctx.FromGroupId
        self.FromUserId: int = ctx.FromUserId
        self.FromNickName: str = ctx.FromNickName
        self.Content: str = ctx.Content
        self.MsgType: str = ctx.MsgType
        self.MsgTime: int = ctx.MsgTime
        self.MsgSeq: int = ctx.MsgSeq
        self.MsgRandom: int = ctx.MsgRandom
        self.RedBaginfo: dict = ctx.RedBaginfo


class _VoiceGroupMsg(_GroupMsg):
    """群语音消息"""

    def __init__(self, ctx: GroupMsg):
        voice_data = json.loads(ctx.Content)
        self.VoiceUrl: str = voice_data['Url']
        self.Tips: str = voice_data['Tips']
        super()._carry_properties(ctx)


class _VideoGroupMsg(_GroupMsg):
    """群视频消息"""

    def __init__(self, ctx: GroupMsg):
        video_data = json.loads(ctx.Content)
        self.ForwordBuf: str = video_data['ForwordBuf']
        self.ForwordField: int = video_data['ForwordField']
        self.Tips: str = video_data['Tips']
        self.VideoMd5: str = video_data['VideoMd5']
        self.VideoSize: str = video_data['VideoSize']
        self.VideoUrl: str = video_data['VideoUrl']
        super()._carry_properties(ctx)


class _GroupPic:
    def __init__(self, pic: dict):
        '''[{"FileId":2161733733,"FileMd5":"","FileSize":449416,"ForwordBuf":"","ForwordField":8,"Url":""}'''
        self.FileId: int = pic.get('FileId')
        self.FileMd5: str = pic.get('FileMd5')
        self.FileSize: int = pic.get('FileSize')
        self.ForwordBuf: str = pic.get('ForwordBuf')
        self.ForwordField: int = pic.get('ForwordField')
        self.Url: str = pic.get('Url')


class _PicGroupMsg(_GroupMsg):
    """群图片/表情包消息"""

    def __init__(self, ctx: GroupMsg):
        pic_data = json.loads(ctx.Content)
        self.GroupPic: List[_GroupPic] = [_GroupPic(i) for i in pic_data['GroupPic']]
        self.Tips: str = pic_data['Tips']
        super()._carry_properties(ctx)
        self.Content: str = pic_data.get('Content')


class _AtGroupMsg(_GroupMsg):
    def __init__(self, ctx: GroupMsg):
        super()._carry_properties(ctx)


class _RedBagGroupMsg(_GroupMsg):
    """群红包消息"""

    def __init__(self, ctx: GroupMsg):
        redbag_info = ctx.RedBaginfo
        self.RedBag_Authkey: str = redbag_info.get('Authkey')
        self.RedBag_Channel: int = redbag_info.get('Channel')
        self.RedBag_Des: str = redbag_info.get('Des')
        self.RedBag_FromType: int = redbag_info.get('FromType')
        self.RedBag_FromUin: int = redbag_info.get('FromUin')
        self.RedBag_Listid: str = redbag_info.get('Listid')
        self.RedBag_RedType: int = redbag_info.get('RedType')
        self.RedBag_StingIndex: str = redbag_info.get('StingIndex')
        self.RedBag_Tittle: str = redbag_info.get('Tittle')
        self.RedBag_Token_17_2: str = redbag_info.get('Token_17_2')
        self.RedBag_Token_17_3: str = redbag_info.get('Token_17_3')
        super()._carry_properties(ctx)


@copy_ctx
def refine_voice_group_msg(ctx: GroupMsg) -> _VoiceGroupMsg:
    """群语音消息"""
    if not isinstance(ctx, GroupMsg):
        raise ContextTypeError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.VoiceMsg:
        return _VoiceGroupMsg(ctx)
    return None


@copy_ctx
def refine_video_group_msg(ctx: GroupMsg) -> _VideoGroupMsg:
    """群视频消息"""
    if not isinstance(ctx, GroupMsg):
        raise ContextTypeError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.VideoMsg:
        return _VideoGroupMsg(ctx)
    return None


@copy_ctx
def refine_pic_group_msg(ctx: GroupMsg) -> _PicGroupMsg:
    """群图片/表情包消息"""
    if not isinstance(ctx, GroupMsg):
        raise ContextTypeError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.PicMsg:
        return _PicGroupMsg(ctx)
    return None


@copy_ctx
def refine_RedBag_group_msg(ctx: GroupMsg) -> _RedBagGroupMsg:
    """群红包消息"""
    if not isinstance(ctx, GroupMsg):
        raise ContextTypeError('Expected `GroupMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.RedBagMsg:
        return _RedBagGroupMsg(ctx)
    return None


#############################Group end##########################################


#############################Friend start#######################################
class _FriendMsg(FriendMsg):
    def _carry_properties(self, ctx: FriendMsg):
        self.message = ctx.message
        self.CurrentQQ = ctx.CurrentQQ

        self.data = ctx.data

        self.FromUin: int = ctx.FromUin
        self.ToUin: int = ctx.ToUin
        self.MsgType: str = ctx.MsgType
        self.MsgSeq: int = ctx.MsgSeq
        self.Content: str = ctx.Content
        self.RedBaginfo: dict = ctx.RedBaginfo


class _VoiceFriendMsg(_FriendMsg):
    """好友语音消息"""

    def __init__(self, ctx: FriendMsg):
        voice_data = json.loads(ctx.Content)
        self.VoiceUrl: str = voice_data['Url']
        self.Tips: str = voice_data['Tips']
        super()._carry_properties(ctx)


class _VideoFriendMsg(_FriendMsg):
    """好友视频消息"""

    def __init__(self, ctx: FriendMsg):
        video_data = json.loads(ctx.Content)
        self.ForwordBuf: str = video_data['ForwordBuf']
        self.ForwordField: int = video_data['ForwordField']
        self.Tips: str = video_data['Tips']
        self.VideoMd5: str = video_data['VideoMd5']
        self.VideoSize: str = video_data['VideoSize']
        self.VideoUrl: str = video_data['VideoUrl']
        super()._carry_properties(ctx)


class _FriendPic:
    def __init__(self, pic: dict):
        """好友图片单个图片所包含的数据
        [{"FileMd5":"","FileSize":0,"Path":"","Url":""}]中的一个
        """
        self.FileMd5: str = pic.get('FileMd5')
        self.FileSize: int = pic.get('FileSize')
        self.Path: str = pic.get('Path')
        self.Url: str = pic.get('Url')


class _PicFriendMsg(_FriendMsg):
    """好友图片/表情包消息"""

    def __init__(self, ctx: FriendMsg):
        pic_data = json.loads(ctx.Content)
        self.FriendPic: List[_FriendPic] = [
            _FriendPic(i) for i in pic_data['FriendPic']
        ]
        self.Tips: str = pic_data['Tips']
        super()._carry_properties(ctx)
        self.Content = pic_data.get('Content')


class _RedBagFriendMsg(_FriendMsg):
    """好友红包消息"""

    def __init__(self, ctx: FriendMsg):
        redbag_info = ctx.RedBaginfo
        self.RedBag_Authkey: str = redbag_info.get('Authkey')
        self.RedBag_Channel: int = redbag_info.get('Channel')
        self.RedBag_Des: str = redbag_info.get('Des')
        self.RedBag_FromType: int = redbag_info.get('FromType')
        self.RedBag_FromUin: int = redbag_info.get('FromUin')
        self.RedBag_Listid: str = redbag_info.get('Listid')
        self.RedBag_RedType: int = redbag_info.get('RedType')
        self.RedBag_StingIndex: str = redbag_info.get('StingIndex')
        self.RedBag_Tittle: str = redbag_info.get('Tittle')
        self.RedBag_Token_17_2: str = redbag_info.get('Token_17_2')
        self.RedBag_Token_17_3: str = redbag_info.get('Token_17_3')
        super()._carry_properties(ctx)


@copy_ctx
def refine_voice_friend_msg(ctx: FriendMsg) -> _VoiceFriendMsg:
    """好友语音消息"""
    if not isinstance(ctx, FriendMsg):
        raise ContextTypeError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.VoiceMsg:
        return _VoiceFriendMsg(ctx)
    return None


@copy_ctx
def refine_video_friend_msg(ctx: FriendMsg) -> _VideoFriendMsg:
    """好友视频消息"""
    if not isinstance(ctx, FriendMsg):
        raise ContextTypeError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.VideoMsg:
        return _VideoFriendMsg(ctx)
    return None


@copy_ctx
def refine_pic_friend_msg(ctx: FriendMsg) -> _PicFriendMsg:
    """好友图片/表情包消息"""
    if not isinstance(ctx, FriendMsg):
        raise ContextTypeError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.PicMsg:
        return _PicFriendMsg(ctx)
    return None


@copy_ctx
def refine_RedBag_friend_msg(ctx: FriendMsg) -> _RedBagFriendMsg:
    """好友红包消息"""
    if not isinstance(ctx, FriendMsg):
        raise ContextTypeError('Expected `FriendMsg`, but got `%s`' % ctx.__class__)
    if ctx.MsgType == MsgTypes.RedBagMsg:
        return _RedBagFriendMsg(ctx)
    return None


#############################Friend end#########################################



import copy
import sys
import time
from collections.abc import Sequence
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, List, Union

import socketio
from schedule import Scheduler as _Scheduler

from .config import config
from .logger import logger
from .model import EventMsg, FriendMsg, GroupMsg, model_map
from .plugin import PluginManager
from .typing import EventMsgReceiver, FriendMsgReceiver, GroupMsgReceiver


def _deco_creater(bind_type):
    def deco(self, func):
        if bind_type == 'OnGroupMsgs':
            self.add_group_msg_receiver(func)
        elif bind_type == 'OnFriendMsgs':
            self.add_friend_msg_receiver(func)
        else:
            self.add_event_receiver(func)

    return deco


class IOTBOT:  # pylint: disable = too-many-instance-attributes
    """
    :param qq: 机器人QQ号(多Q就传qq号列表)
    :param use_plugins: 是否开启插件功能
    :param plugin_dir: 插件存放目录
    :param group_blacklist: 群黑名单, 此名单中的群聊消息不会被处理,默认为空，即全部处理
    :param friend_whitelist: 好友白名单，只有此名单中的好友消息才会被处理，默认为空，即全部处理
    :param log: 是否开启日志
    :param log_file: 是否文件日志
    :param port: 运行端口
    :param beat_delay: 心跳延时时间（s）
    :param host: ip，需要包含schema
    """

    def __init__(
        self,
        qq: Union[int, List[int]],
        use_plugins: bool = False,
        plugin_dir: str = 'plugins',
        group_blacklist: List[int] = None,
        friend_blacklist: List[int] = None,
        log: bool = True,
        log_file: bool = True,
        port: int = 8888,
        host: str = 'http://127.0.0.1',
    ):
        if isinstance(qq, Sequence):
            self.qq = list(qq)
        else:
            self.qq = [qq]
        self.use_plugins = use_plugins
        self.plugin_dir = plugin_dir
        self.host = config.host or host
        self.port = config.port or port
        self.group_blacklist = set(config.group_blacklist or group_blacklist or [])
        self.friend_blacklist = set(config.friend_blacklist or friend_blacklist or [])

        # 作为程序是否应该退出的标志，以便后续用到，如定时任务
        self._exit = False

        if log:
            if log_file:
                logger.add(
                    './logs/{time}.log',
                    format='{time:YYYY-MM-DD HH:mm} {level}\t{message}',
                    rotation='1 day',
                    encoding='utf-8',
                )
        else:
            logger.disable(__name__)

        # 用于定时任务
        self.scheduler = _Scheduler()

        # 手动添加的消息接收函数
        self.__friend_msg_receivers_from_hand = []
        self.__group_msg_receivers_from_hand = []
        self.__event_receivers_from_hand = []

        # webhook 里的消息接收函数，是特例
        if config.webhook:
            from . import webhook  # pylint:disable=import-outside-toplevel

            # 直接加载进 `hand`
            self.__friend_msg_receivers_from_hand.append(webhook.receive_friend_msg)
            self.__group_msg_receivers_from_hand.append(webhook.receive_group_msg)
            self.__event_receivers_from_hand.append(webhook.receive_events)

        # 消息上下文对象中间件
        self.__friend_context_middleware: Callable[[FriendMsg], FriendMsg] = None
        self.__group_context_middleware: Callable[[GroupMsg], GroupMsg] = None
        self.__event_context_middleware: Callable[[EventMsg], EventMsg] = None

        # 插件管理
        self.plugMgr = PluginManager(self.plugin_dir)
        if use_plugins:
            self.plugMgr.load_plugins()
            print(self.plugin_status)

        # 依次各种初始化
        self.__initialize_socketio()
        self.__refresh_executor()
        self.__initialize_handlers()

    ########################################################################
    # shortcuts to call plugin manager methods
    ########################################################################
    # 只推荐使用这几个方法，其他的更细致的方法需要通过 plugMgr 对象访问
    def load_plugins(self):
        '''加载新插件'''
        self.plugMgr.load_plugins()

    def reload_plugins(self):
        '''重载旧插件，加载新插件'''
        self.plugMgr.reload_plugins()

    def reload_plugin(self, plugin_name: str):
        '''重载指定插件'''
        self.plugMgr.reload_plugin(plugin_name)

    def refresh_plugins(self):
        '''刷新插件目录所有插件'''
        self.plugMgr.refresh()

    def remove_plugin(self, plugin_name: str):
        '''停用指定插件'''
        self.plugMgr.remove_plugin(plugin_name)

    def recover_plugin(self, plugin_name: str):
        '''启用指定插件'''
        self.plugMgr.recover_plugin(plugin_name)

    @property
    def plugin_status(self):
        '''插件启用状态'''
        return self.plugMgr.info_table

    @property
    def plugins(self):
        '''插件名列表'''
        return self.plugMgr.plugins

    @property
    def removed_plugins(self):
        '''已停用的插件名列表'''
        return self.plugMgr.removed_plugins

    ########################################################################

    # 调度定时任务
    def _run_padding(self):
        logger.info(f'{len(self.scheduler.jobs)} tasks that are scheduled to run.')
        while True:
            # 定时任务不是主角，所以必须其他工作正常才运行
            if self._exit:
                return
            s = self.scheduler
            runnable_jobs = (job for job in s.jobs if job.should_run)
            for job in sorted(runnable_jobs):
                if hasattr(job.job_func, '__name__'):
                    job_func_name = job.job_func.__name__
                else:
                    job_func_name = repr(job.job_func)
                logger.info('Running task => %s' % job_func_name)
                s._run_job(job)  # pylint: disable=protected-access
            # 避免不必要的循环
            # 因为定时任务间隔都很长，进入延时后并不能很好的判断self._exit来退出
            # 所以这里每次只延时`delay`秒就判断一次是否应该退出
            delay = 5
            t = int(s.idle_seconds - 5)  # 延时不是必要的，减去5s抵消下面的时间消耗
            if t <= 0:
                continue
            for _ in range(t // delay):
                if self._exit:
                    return
                time.sleep(delay)

    def run(self):
        logger.info('Connecting to the server...')
        try:
            self.socketio.connect(f'{self.host}:{self.port}', transports=['websocket'])
        except Exception:
            logger.exception('连接失败')
            self.__executor.shutdown(False)
            self._exit = True
            sys.exit(1)
        else:
            try:
                self.socketio.wait()
            except KeyboardInterrupt:
                self.__executor.shutdown(wait=False)
                self._exit = True
                print('\nbye~')
                sys.exit(0)

    def connect(self):
        logger.success('Connected to server successfully!')

        # GetWebConn
        for qq in self.qq:
            self.socketio.emit(
                'GetWebConn',
                str(qq),
                callback=lambda x: logger.info(
                    f'GetWebConn -> {qq} => {x}'  # pylint: disable=cell-var-from-loop
                ),
            )

        # 启动定时任务
        if len(self.scheduler.jobs) != 0:
            self.__executor.submit(self._run_padding).add_done_callback(
                self.__thread_pool_callback
            )

    @property
    def receivers(self):
        '''消息处理函数数量'''
        return {
            'friend': len(
                (
                    *self.plugMgr.friend_msg_receivers,
                    *self.__friend_msg_receivers_from_hand,
                )
            ),
            'group': len(
                (
                    *self.plugMgr.group_msg_receivers,
                    *self.__group_msg_receivers_from_hand,
                )
            ),
            'event': len(
                (*self.plugMgr.event_receivers, *self.__event_receivers_from_hand)
            ),
        }

    @receivers.setter
    def receivers(self, _):
        logger.warning('The attribute receivers is read-only!')

    def __initialize_socketio(self):
        self.socketio = socketio.Client()
        self.socketio.event()(self.connect)

    def __refresh_executor(self):
        # 根据消息接收函数数量初始化线程池
        self.__executor = ThreadPoolExecutor(
            max_workers=min(
                50,
                len(
                    [  # 减小数量，控制消息频率
                        *self.plugMgr.friend_msg_receivers,
                        *self.__friend_msg_receivers_from_hand,
                        *self.plugMgr.group_msg_receivers,
                        *self.__group_msg_receivers_from_hand,
                        *self.plugMgr.event_receivers,
                        *self.__event_receivers_from_hand,
                        *range(3),
                    ]
                )
                * 2,
            )
        )

    # 手动添加
    def add_group_msg_receiver(self, func: GroupMsgReceiver):
        '''添加群消息接收函数'''
        self.__group_msg_receivers_from_hand.append(func)

    def add_friend_msg_receiver(self, func: FriendMsgReceiver):
        '''添加好友消息接收函数'''
        self.__friend_msg_receivers_from_hand.append(func)

    def add_event_receiver(self, func: EventMsgReceiver):
        '''添加事件消息接收函数'''
        self.__event_receivers_from_hand.append(func)

    ########################################################################
    # context distributor
    ########################################################################
    def __friend_context_distributor(self, context: FriendMsg):
        for f_receiver in [
            *self.__friend_msg_receivers_from_hand,
            *self.plugMgr.friend_msg_receivers,
        ]:
            self.__executor.submit(
                f_receiver, copy.deepcopy(context)
            ).add_done_callback(self.__thread_pool_callback)

    def __group_context_distributor(self, context: GroupMsg):
        for g_receiver in [
            *self.__group_msg_receivers_from_hand,
            *self.plugMgr.group_msg_receivers,
        ]:
            self.__executor.submit(
                g_receiver, copy.deepcopy(context)
            ).add_done_callback(self.__thread_pool_callback)

    def __event_context_distributor(self, context: EventMsg):
        for e_receiver in [
            *self.__event_receivers_from_hand,
            *self.plugMgr.event_receivers,
        ]:
            self.__executor.submit(
                e_receiver, copy.deepcopy(context)
            ).add_done_callback(self.__thread_pool_callback)

    ########################################################################
    # register context middleware
    ########################################################################
    def register_friend_context_middleware(
        self, middleware: Callable[[FriendMsg], FriendMsg]
    ):
        """注册好友消息中间件"""
        if self.__friend_context_middleware is not None:
            raise Exception('Cannot register more than one middleware(friend)')
        self.__friend_context_middleware = middleware

    def register_group_context_middleware(
        self, middleware: Callable[[GroupMsg], GroupMsg]
    ):
        """注册群消息中间件"""
        if self.__group_context_middleware is not None:
            raise Exception('Cannot register more than one middleware(group)')
        self.__group_context_middleware = middleware

    def register_event_context_middleware(
        self, middleware: Callable[[EventMsg], EventMsg]
    ):
        """注册事件消息中间件"""
        if self.__event_context_middleware is not None:
            raise Exception('Cannot register more than one middleware(event)')
        self.__event_context_middleware = middleware

    ########################################################################
    # message handler
    ########################################################################
    @logger.catch
    def __thread_pool_callback(self, worker):
        worker_exception = worker.exception()
        if worker_exception:
            raise worker_exception

    def __friend_msg_handler(self, msg):
        context: FriendMsg = model_map['OnFriendMsgs'](msg)
        logger.info(f'{context.__class__.__name__} ->  {context.data}')
        # 黑名单
        if context.FromUin in self.friend_blacklist:
            return
        # 中间件
        if self.__friend_context_middleware is not None:
            new_context = self.__friend_context_middleware(context)
            if isinstance(new_context, type(context)):
                context = new_context
        self.__executor.submit(self.__friend_context_distributor, context)

    def __group_msg_handler(self, msg):
        context: GroupMsg = model_map['OnGroupMsgs'](msg)
        logger.info(f'{context.__class__.__name__} ->  {context.data}')
        # 黑名单
        if context.FromGroupId in self.group_blacklist:
            return
        # 中间件
        if self.__group_context_middleware is not None:
            new_context = self.__group_context_middleware(context)
            if isinstance(new_context, type(context)):
                context = new_context
        self.__executor.submit(self.__group_context_distributor, context)

    def __event_msg_handler(self, msg):
        context: EventMsg = model_map['OnEvents'](msg)
        logger.info(f'{context.__class__.__name__} ->  {context.data}')
        # 中间件
        if self.__event_context_middleware is not None:
            new_context = self.__event_context_middleware(context)
            if isinstance(new_context, type(context)):
                context = new_context
        self.__executor.submit(self.__event_context_distributor, context)

    def __initialize_handlers(self):
        self.socketio.on('OnGroupMsgs')(self.__group_msg_handler)
        self.socketio.on('OnFriendMsgs')(self.__friend_msg_handler)
        self.socketio.on('OnEvents')(self.__event_msg_handler)

    ###########################################################################
    # decorators
    on_group_msg = _deco_creater('OnGroupMsgs')
    on_friend_msg = _deco_creater('OnFriendMsgs')
    on_event = _deco_creater('OnEvents')

    def __repr__(self):
        return 'IOTBOT <{}> <host-{}> <port-{}>'.format(
            " ".join([str(i) for i in self.qq]), self.host, self.port
        )



"""Author: wongxy

github.com/xiyaowong/python-iotbot

>>> from iotbot import IOTBOT, GroupMsg, FriendMsg, EventMsg

>>> bot = IOTBOT(your_bot_qq)

>>> @bot.on_group_msg
>>> def group(ctx: GroupMsg):
>>>    pass

>>> @bot.on_friend_msg
>>> def friend(ctx: FriendMsg):
>>>    pass

>>> @bot.on_event
>>> def event(message: EventMsg):
>>>     pass

>>> if __name__ == '__main__':
        bot.run()
"""

from . import refine as refine_message  # 兼容旧版本，以后可能删除
from .action import Action
from .client import IOTBOT
from .model import EventMsg, FriendMsg, GroupMsg

__author__ = 'wongxy'



from typing import Any, Callable

from .model import EventMsg, FriendMsg, GroupMsg

GroupMsgReceiver = Callable[[GroupMsg], Any]
FriendMsgReceiver = Callable[[FriendMsg], Any]
EventMsgReceiver = Callable[[EventMsg], Any]



from iotbot import IOTBOT, Action, EventMsg, FriendMsg, GroupMsg

bot_qq = 11
bot = IOTBOT(bot_qq, use_plugins=False)
action = Action(bot)


@bot.on_group_msg
def on_group_msg(ctx: GroupMsg):
    print(ctx.message)
    if ctx.Content == '刷新插件':
        bot.refresh_plugins()
    if ctx.Content == '.test':
        action.send_group_pic_msg(ctx.FromGroupId, 'http://url.cn/cqjYvc06')


@bot.on_friend_msg
def on_friend_msg(ctx: FriendMsg):
    print(ctx.message)
    if ctx.Content == '.test':
        action.send_friend_pic_msg(ctx.FromUin, picUrl='http://url.cn/cqjYvc06')


@bot.on_event
def on_event(ctx: EventMsg):
    print(ctx.message)
    print(ctx.EventName)


def test_group(ctx: GroupMsg):
    print('In test_group', ctx.FromNickName)


bot.add_group_msg_receiver(test_group)

if __name__ == "__main__":
    print(bot.receivers)
    bot.run()
